package SketchAggSensor.topology;

/**
 * <p>Title: Aggregation Query in Sensor Network with Sketches</p>
 * <p>Description: This code is for the simulation of sensor network with sketches algorithms. It applies with aggregation queries.</p>
 * <p>Copyright: Computer Science Department, Boston University,Copyright (c) 2003</p>
 * <p>Company: Boston University</p>
 * @author Feifei Li
 * @version 1.0
 */

import java.util.*;
import SketchAggSensor.*;
import SketchAggSensor.sensor.*;
import SketchAggSensor.message.*;

public class Topology implements ClockListener {
    static final double BASE_READING = 0d;
    static final double MAX_READING = 1000d;
    public static final int BYTES_PER_MESSAGE = 30;

    Vector[][] sensors;
    int xdim, ydim;

    //sensor list keeps all the sensors in a single least
    //for easy statistic computation
    Vector sensorList;

    Vector msgQueue = new Vector();
    Vector nextEpochMsgQueue = new Vector();
    boolean sending = false;
    int epoch = 0;

    protected int sent = 0; //total messages sent through the topology
    protected int sentBytes = 0; //total bytes through the topology
    protected int comm_range = 1; //communications radius

    //mapping from id :: SensorIntf
    private HashMap sensorsHash = new HashMap();

    /** Constructor -- intiailize an empty topology
        of the specified size.

        @param width The width of the topology, in integer units
        @param height The height of the topology
    */
    public Topology(int width, int height, int radius) {
        xdim = width;
        ydim = height;

        sensors = new Vector[width][height];
        sensorList = new Vector();
        comm_range = radius;
    }


    /** Add a sensor to the topology at the specified location
        The (internally stored) location of the sensor is also updated
        @param s The sensor to add
        @param x The x location of the sensor
        @param y The y location of the sensor
    */
    public void addSensor(SensorInterface s, int x, int y) {
        Vector v = sensors[x][y];
        if (v == null) {
            v = new Vector();
            sensors[x][y] = v;
        }

        v.addElement(s);
        sensorList.addElement(s);
        try {
            s.setX(x);
            s.setY(y);
        } catch (Exception e) {
            System.out.println("Couldn't set loc to (" + x + "," + y + ")");
        }
        sensorsHash.put(new Integer(s.getId()), s);
    }


    public int numSensors() {
        return sensorsHash.size();
    }

    /** @return Vector containing the sensors at (x,y) (see {@link Sensor Sensor})
     */
    public Vector getSensorsAt(int x, int y) {
        return sensors[x][y];
    }

    public SensorInterface getSensorId(int id) {
        return (SensorInterface)(sensorsHash.get(new Integer(id)));
    }

    /** Radiate a message out from the specified point on the
        topology.  Message radiate 1 unit in all directions.
        Messages are not echoed to their sender .

        Handles reentrancy -- so sensors that send messages while
        a message is being delivered will have their messages enqueued.

        @param m The message to send
        @param x The x coordinate of the origin of the message
        @param y The y coordinate of the origin of the message
    */
    public void radiateMessage(Message m, int x, int y) {
        while (true) {
            Enumeration e;


        //for now, assume a simple connectivity graph
        //where anyone within one square of the message
        //origin hears the message

        //look at the senders clock, not the message's time stamp
            if (sending || epoch < Clock.getEpoch(m.getSender().getTime())) {
                Vector msgEl = new Vector();
                msgEl.addElement(m);
                msgEl.addElement(new Integer(x));
                msgEl.addElement(new Integer(y));
                if (epoch ==  Clock.getEpoch(m.getSender().getTime()))
                    msgQueue.addElement(msgEl);
                else
                    nextEpochMsgQueue.addElement(msgEl);
                return;
            }
            sending = true;
            //kind of a hack -- so that underlying sensors don't have to worry
            //about message size, charge cost of oversize messages here
            int thisRound;
            thisRound = (m.numBytes() / BYTES_PER_MESSAGE);
            if (thisRound * BYTES_PER_MESSAGE < m.numBytes() || m.numBytes() == 0)
              thisRound++;
            sent += thisRound;
            sentBytes += m.numBytes();
            e = neighbors(x,y);
            while (e.hasMoreElements()) {
                Sensor s = (Sensor)e.nextElement();
                if (s != m.getSender()) {
                    s.receiveMessage(m);
                }
            }
            //add by feifei
            /*Vector recipients = m.getRecipients();
            for (int i = 0; i < recipients.size(); i++)
              //if (((Integer)recipients.elementAt(i)).intValue() == id)
            {
              Sensor s =(Sensor) this.getSensorId(i);
              s.receiveMessage(m);
            }*/

            sending = false;
            if (msgQueue.size() == 0) {
                break;
            } else {
                //if a message was enqueued while sending the previous message, dequeue it and
                //send it out...
                Vector msgEl = (Vector)msgQueue.remove(0);
                //no tail recursion in java, so just overwrite params and loop
                m = (Message)msgEl.elementAt(0);
                x = ((Integer)msgEl.elementAt(1)).intValue();
                y = ((Integer)msgEl.elementAt(2)).intValue();

            }
        }
    }


    /** Return an enumeration of neighbors (type {@link Sensor Sensor})
     Allows extensions which create funky connectivities */
    public Enumeration neighbors(int x, int y) {
        Vector neighbors = new Vector(),v;

        for (int i = x-comm_range; i <= x+comm_range; i++) {
            for (int j = y - comm_range; j <= y + comm_range; j++) {
                if (validLoc(i,j)) {
                    v = sensors[i][j];
                    if (v != null) neighbors.addAll(v);
                }
            }

        }
        return neighbors.elements();
    }

    /** @return true iff the specified coordinates are a valid
        location within the topology
    */
    public boolean validLoc(int x, int y) {
        return (x >= 0 && x < xdim &&
                y >= 0 && y < ydim);
    }


    public boolean validXLoc(int x) {
        return (x >= 0 && x < xdim );
    }


    public boolean validYLoc( int y) {
        return (y >= 0 && y < ydim);
    }

    /** @return The width, in sensors of the topology */
    public int width() {
        return xdim;
    }

    /** @return The height, in sensors of the topology */
    public int height() {
        return ydim;
    }

    /** @return An enumeration of all sensors in this topology */
    public Enumeration sensors() {
        return sensorList.elements();
    }

    /** @return A count of the total number of messages sent in this topology */
    public int sentCount() {
        return sent;
    }

    /** @return A count of the total number of bytes sent in this topology */
    public int sentBytes() {
        return sentBytes;
    }

    /** @return A count of the total number of messages heard by all sensors in this topology */
    public int readCount() {
        Enumeration e = sensors();
        int cnt = 0;

        while (e.hasMoreElements()) {
            cnt += ((Sensor)e.nextElement()).msgCount();
        }
        return cnt;
    }

    /** Disable the sensors at the specified location.
        Disabled sensors neither send for forward aggregate messages
    */
    public void killSensors(int x, int y) {
        Enumeration e = sensors[x][y].elements();
        while (e.hasMoreElements()) {
            ((Sensor)e.nextElement()).setDisabled(true);
        }
    }

    /** Deactivate the sensors at the specified location.
        Deactivated sensors forward aggregate values, but don't include their
        own readings.
    */
    public void deactivateSensors(int x, int y) {
        Enumeration e = sensors[x][y].elements();
        while (e.hasMoreElements()) {
            ((Sensor)e.nextElement()).setNoParticipate(true);
        }
    }

    /** Return the sensor reading at the current location... */
    public int getReading(int x, int y) {
        double dx = ((MAX_READING - BASE_READING)/2) / (double)xdim;
        double dy = ((MAX_READING - BASE_READING)/2) / (double)ydim;

        return (int)(dx * x + dy * y);

    }


    /* Send out any messages that arrived during the last epoch to be delivered
       in this epoch ...
    */
    public void setTime(int time) {
        int lastEpoch = epoch;

        epoch = Clock.getEpoch(time);

        if (epoch > lastEpoch) {
            msgQueue = nextEpochMsgQueue;
            nextEpochMsgQueue = new Vector();
            if (msgQueue.size() > 0) {
                Vector v = (Vector)msgQueue.remove(0);
                radiateMessage((Message)v.elementAt(0), ((Integer)v.elementAt(1)).intValue(),
                               ((Integer)v.elementAt(2)).intValue());
            }
        }
    }


}
